unit unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, ExtCtrls, LCLType, LclIntf, StdCtrls,
  OpenGLContext, GL, math;

type

  { TForm1 }

  TForm1 = class(TForm)
    Label1: TLabel;
    OpenGLControl1: TOpenGLControl;
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure OpenGLControl1Paint(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    FRotation: GLfloat;
    procedure InitGL;
  end;

var
  Form1: TForm1;

const
  LightDiffuse: array[0..3] of Single = (1, 1, 1, 1);

implementation

type
  TVector3f = record
    x, y, z: GLfloat;
  end;

  TColor3f = record
    R, G, B: GLfloat;
  end;

function Vector3f(x, y, z: GLfloat): TVector3f;
begin
  Result.x := x;
  Result.y := y;
  Result.z := z;
end;

operator - (a, b: TVector3f): TVector3f;
begin
  Result := Vector3f(a.x - b.x, a.y - b.y, a.z - b.z);
end;

operator / (a: TVector3f; b: GLfloat): TVector3f;
begin
  Result := Vector3f(a.x/b, a.y/b, a.z/b);
end;

function Cross(a, b: TVector3f): TVector3f;
begin
  Result.x := a.y * b.z - a.z * b.y;
  Result.y := a.x * b.z - a.z * b.x;
  Result.z := a.x * b.y - a.y * b.x;
end;

function VLen(a: TVector3f): Double;
begin
  Result := sqrt(sqr(a.x) + sqr(a.y) + sqr(a.z));
end;

function Normalize(a: TVector3f): TVector3f;
begin
  Result := a / VLen(a);
end;

// P1, P2, P3 oriented according to right-hand rule for normal pointing outward
function Normal(P1, P2, P3: TVector3f): TVector3f;
var
  v1, v2: TVector3f;
begin
  v1 := P2 - P1;
  v2 := P3 - P1;
  Result := Normalize(Cross(v1, v2));
end;

procedure gluPerspective(fovy, aspect, zNear, zFar: GLdouble);
var
  Width, Height: Double;
begin
  Height := tan(fovy * pi / 360) * zNear;
  Width  := Height * aspect;
  glFrustum(-Width, Width, -Height, Height, zNear, zFar);
end;

{$R *.lfm}

{ TForm1 }

procedure Triangle(P1, P2, P3: TVector3f; C1, C2, C3: TColor3f);
var
  n: TVector3f;
begin
  n := Normal(P1, P2, P3);
  glNormal3f(n.x, n.y, n.z);

  glColor3f(C1.R, C1.G, C1.B); glVertex3f(P1.x, P1.y, P1.z);
  glColor3f(C2.R, C2.G, C2.B); glVertex3f(P2.x, P2.y, P2.z);
  glColor3f(C3.R, C3.G, C3.B); glVertex3f(P3.x, P3.y, P3.z);
end;

{ Draws a pyramid defined by the corners of a square (points A, B, C, D) and
  the apex E above the center of the square.

  Top view:

     D --------C
     |  \   /  |
     |    E    |          y ---> x     y pointing outside the screen in this view
     |  /   \  |          |
     A --------B          v
                          z
  There are triangles
    ABE, BCE, CDE, DAE
    (points ordered so that the normal points outward if the right-hand-rule is applied. }

procedure TForm1.OpenGLControl1Paint(Sender: TObject);
const
  AlreadyInit: Boolean = False;
  A: TVector3f = (x:-1; y:-1; z:+1);
  B: TVector3f = (x:+1; y:-1; z:+1);
  C: TVector3f = (x:+1; y:-1; z:-1);
  D: TVector3f = (x:-1; y:-1; z:-1);
  E: TVector3f = (x: 0; y:+1; z: 0);

  COLOR_R: TColor3f = (R:1.0; G:0.0; B:0.0);  // red
  COLOR_G: TColor3f = (R:0.0; G:1.0; B:0.0);  // green
  COLOR_B: TColor3f = (R:0.0; G:0.0; B:1.0);  // blue
begin
  if not AlreadyInit then
  begin
    InitGL;
    AlreadyInit := true;
  end;

  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glLoadIdentity;
  glTranslatef(0, 0, -5);
  glRotatef(FRotation, 0, 1, 0);
  glBegin(GL_TRIANGLES);
    Triangle(A, B, E, COLOR_R, COLOR_G, COLOR_B);
    Triangle(B, C, E, COLOR_G, COLOR_R, COLOR_B);
    Triangle(C, D, E, COLOR_R, COLOR_G, COLOR_B);
    Triangle(D, A, E, COLOR_G, COLOR_R, COLOR_B);
  glEnd;

  OpenGLControl1.SwapBuffers;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  OpenGLControl1.Anchors := [akTop, akLeft, akBottom, akRight];
  OpenGLControl1.MultiSampling := 8;
  FRotation := 0;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  OpenGLControl1.Invalidate;

  if GetKeyState(VK_RSHIFT) < 0 then
    FRotation := FRotation - 0.8
  else
    if GetKeyState(VK_LSHIFT) < 0 then
      FRotation := FRotation + 0.8;
end;

procedure TForm1.InitGL;
begin
  glShadeModel(GL_SMOOTH);
  glClearColor(0, 0, 0, 0);
  glClearDepth(1);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  glViewport(0, 0, OpenGLControl1.Width, OpenGLControl1.Height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(45, OpenGLControl1.Width/OpenGLControl1.Height, 0.1, 100);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;

  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);
  glEnable(GL_LIGHT1);
end;

end.
